/*
 *      OFFICIAL NOTIFICATION: the following CONFIDENTIAL and PROPRIETARY 
 * 	property legend shall not be removed from this source code module 
 * 	for any reason.
 *
 *	This program is the CONFIDENTIAL and PROPRIETARY property 
 *	of FairCom(R) Corporation. Any unauthorized use, reproduction or
 *	transfer of this computer program is strictly prohibited.
 *
 *      Copyright (c) 1984 - 2000 FairCom Corporation.
 *	This is an unpublished work, and is subject to limited distribution and
 *	restricted disclosure only. ALL RIGHTS RESERVED.
 *
 *			RESTRICTED RIGHTS LEGEND
 *	Use, duplication, or disclosure by the Government is subject to
 *	restrictions set forth in subparagraph (c)(1)(ii) of the Rights in
 * 	Technical Data and Computer Software clause at DFARS 252.227-7013.
 *	FairCom Corporation, 2100 Forum Blvd. C, Columbia, MO 65203.
 *
 *	c-tree PLUS(tm)	Version 6.10
 *			Release B
 *			June 25, 2000
 */

/* Header files. ***********************************************************/

#ifdef PROTOTYPE
#include <stdarg.h>
#else
#include <varargs.h>
#endif


/* Global constants. *******************************************************/

/*
 * Uncomment this define to log times to disk.
#define ctLOGTIMESTODISK
 */

#define SVN_DEFAULT "FAIRCOMS"	/* Default server name.		*/
#define MAXTRD	     8		/* Maximum threads to display.	*/
#define NUML	     6
#define NAML        24
#define CLSL         2
#define PADG        88

#ifndef ctrt_printf
#define ctrt_printf printf /* ensure console (non-GUI) output */
#endif

#ifdef ctPortUNIX
#ifdef ERR
#undef ERR	/* conflict with curses.h */
#endif
#define USE_CURSES
#define ctCURSES_H <curses.h>
#include ctCURSES_H
#ifdef VOID
#undef VOID
#define VOID void
#endif
#else
#define USE_ANSI
#endif

#ifdef USE_ANSI

#ifdef ctPortWNTALPHA
#include <wincon.h>
#define LOC(y,x)                             \
{                                            \
    COORD pos = {x-1, y-1};                  \
    SetConsoleCursorPosition(ghStdOut, pos); \
}
#define CLS()     { int i; for (i=0; i<=MAXY; i++) printf("\n"); }

#else  /* ~ctPortWNTALPHA */

#define LOC(y,x)  printf("%c[%d;%dH", 27, y, x)
#define CLS()     printf("%c[H%c[J", 27, 27)

#endif /* ~ctPortWNTALPHA */

#define MAXY 24
#define MAXX 80
#define WINY (MAXY-2) /* Number of lines for thread output. */
#define Y 0
#define X 1

#endif /* USE_ANSI */


/* Global structures. ******************************************************/

typedef struct tm ANSITIME;

typedef struct		/* Data record structure. */
{
	COUNT dlfg;
	TEXT  unqu[NUML];
	LONG  sern; /* Auto serial number. */
	COUNT dlen;
	TEXT  name[NAML];
	TEXT  clss[CLSL];
	TEXT  grbg[128 - ctSIZE(COUNT) - ctSIZE(LONG) - ctSIZE(COUNT) -
               NUML - NAML - CLSL];
} DREC;

#ifdef WTMTAP_C
typedef struct		/* Window information. */
{
	HWND  hwnd;
	ULONG iLine;
} ctWINCTL;
#endif /* WTMTAP_C */

typedef struct		/* Application options. */
{
	TEXT	action;		/* Type of action.			*/
	TEXT	filestat;	/* Create new files?			*/
	TEXT	mode;		/* Transaction processing mode.		*/
	LONG	trials;		/* Number of trials.			*/
	pTEXT	svn;		/* Server name.				*/
	UCOUNT	seed;		/* Random number seed.			*/
	COUNT	trnlen;		/* Percentage of transaction aborts.	*/
	NINT	maxfiles;	/* Maximum number of files.		*/
	NINT	threads;	/* Number of threads.			*/
	LONG	round;		/* Round number (1, 2, or 3).		*/
} APPOPTS, *pAPPOPTS;


/* Function prototypes. ****************************************************/
#ifdef PROTOTYPE
VOID appDescribe(VOID);
#else
VOID appDescribe();
#endif
#ifdef CTMTAP_C
#ifdef PROTOTYPE
VOID th_printf(NINT t, pTEXT arg0, ... );
#else
VOID th_printf();
#endif
#else  /* ~CTMTAP_C */
VOID th_printf(LONG proc_id, pTEXT fmt, ... );
#endif /* ~CTMTAP_C */


/* Global variables. *******************************************************/

COUNT     trnlen;	/* Percentage of transaction aborts.		*/
UCOUNT    seed;		/* Random number seed.				*/
NINT      tfin;		/* Number of threads which have finished.	*/
NINT      qid;		/* Inter-thread FIFO message queue ID.		*/
ctMUTEX   tfin_mtx;	/* Mutex for thread finished counter.		*/
#if (defined(USE_CURSES) || defined(USE_ANSI))
ctMUTEX   tprt_mtx;	/* Mutex for displaying thread output.		*/
#endif /* USE_CURSES || USE_ANSI */
ANSITIME *ct_tm;	/* Start/end date and time.			*/
TEXT      mode;		/* Transaction processing mode.			*/
TEXT      action;	/* Type of action.				*/
pTEXT     svn;		/* Server name.					*/
ctBLOCK	  guiblk;	/* Block used to wait for window creation.	*/
#ifdef WTMTAP_C
ctWINCTL  ghwnd[MAXTRD];/* Information for each thread's window.	*/
#endif /* WTMTAP_C */
#ifdef TRANPROC
COUNT	  filmod;	/* File mode.					*/
#endif
#ifdef ctLOGTIMESTODISK
FILE	 *fpTIMES;	/* Working file pointer to log info to disk.	*/
#endif

#ifdef USE_CURSES
NINT      width;          /* Display width for each thread. */
WINDOW   *win[MAXTRD];  /* Window list. */
WINDOW   *bdr[MAXTRD];	/* Border list. */
#endif

#ifdef USE_ANSI
#ifdef ctPortWNTALPHA
HANDLE    ghStdOut;
#endif
NINT      width;                /* Display width for each thread. */
COUNT     currpos[MAXTRD][2]; /* Current window position for threads. */
COUNT     buftop[MAXTRD];     /* Top of display in win buffer. */
TEXT      win[WINY][MAXX];      /* Window buffer for all threads. */
#endif


/* ************************************************************************* */
/* ************************************************************************* */

/*
 * sysdate: Return current system time.
 */

#ifdef PROTOTYPE
LONG sysdate(void)
#else
LONG sysdate()
#endif
{
	LONG ltime;

	time((time_t *)&ltime);
	ct_tm = (ANSITIME *)localtime((time_t *)&ltime);
	return(ltime);
} /* sysdate() */


/* ************************************************************************* */
/* ************************************************************************* */

/*
 * rate: Compute and display elapsed time and time per operation for a trial.
 */

#ifdef PROTOTYPE
VOID rate(NINT proc_id,LONG lstop,LONG lstart,LONG trials,pTEXT th_str)
#else
VOID rate(proc_id,lstop,lstart,trials,th_str)
NINT	proc_id;
LONG	lstop,lstart,trials;
pTEXT	th_str;
#endif
{
	if (!trials)
		return;
	lstop -= lstart;        /* Elapsed seconds. */
	th_printf(proc_id,
"Elapsed time: %ld (sec)\nTime per op: %ld (sec/1000)\n",
		lstop, (lstop * 1000) / trials);

#ifdef ctLOGTIMESTODISK
	fprintf(fpTIMES,
"\n%s  Elapsed time : %ld (sec) time per op = %ld (sec/1000)\n\n",
		th_str, lstop, (lstop * 1000) / trials);  
#endif
} /* rate() */


#ifdef TRANPROC
/* ************************************************************************* */
/* ************************************************************************* */

/*
 * tbeg: Start a transaction.
 */

#ifdef PROTOTYPE
COUNT tbeg(pNINT ptlen,pNINT ptrst)
#else
COUNT tbeg(ptlen,ptrst)
pNINT ptlen,ptrst;
#endif
{
	COUNT status = (COUNT)0;

	if (*ptlen == 0)
	{
		*ptlen = trnlen;
		*ptrst = 0;
		TRANBEG((COUNT)(mode == 't' ? (PREIMG | ENABLE) : (TRNLOG | ENABLE)));
		status = uerr_cod;
	}
	return(status);
} /* tbeg() */
#endif


#ifdef TRANPROC
/* ************************************************************************* */
/* ************************************************************************* */

/*
 * tend: Commit or abort a transaction, or set or restore a save point.
 */

#ifdef PROTOTYPE
COUNT tend(NINT proc_id,pNINT ptlen,NINT status,pNINT ptrst,pNINT tcnt,NINT tchg)
#else
COUNT tend(proc_id,ptlen,status,ptrst,tcnt,tchg)
NINT  proc_id;
pNINT ptlen,ptrst,tcnt;
NINT  status,tchg;
#endif
{
	if (status)
		if (*ptlen == trnlen || ++(*ptrst) > 3)
		{
			*ptlen = *ptrst = 0;
			TRANABT();
			return(status);
		}
		else
		{
			TRANRST(-1);
			return(status);
		}

	(*ptlen)--;
	if (*ptlen == 0)
	{
		*ptrst = 0;
		if (status)
			status = TRANABT();
		else
		{
			status = TRANEND(FREE);
			if (status == 0)
			{
				*tcnt += tchg;
				th_printf(proc_id, "\r  TT %d", *tcnt);
			}
		}
		return(status);
	} else
		TRANSAV();

	return(0);
} /* tend() */
#endif


/* ************************************************************************* */
/* ************************************************************************* */

/*
 * unif: Return a number in the range l..u.
 */

#ifdef PROTOTYPE
UCOUNT unif(UCOUNT l,UCOUNT u)
#else
UCOUNT unif(l,u)
UCOUNT      l,u;
#endif
{
	UCOUNT retval;

	seed = 11737 * seed + 3731;
	retval = (seed % ((UCOUNT) (u - l + 1))) + l;

	return(retval);
} /* unif() */


/* ************************************************************************* */
/* ************************************************************************* */

/*
 * setval: Generate a field value.
 */

#ifdef PROTOTYPE
VOID setval(PFAST pTEXT tp,COUNT vl,COUNT fl,COUNT vr)
#else
VOID setval(tp,vl,fl,vr)
PFAST pTEXT tp;
COUNT vl,fl,vr;
#endif
{
	FAST COUNT i;
	COUNT      keyl;
	TEXT       lstchr;

	for (i = 0; i < vl; i++)
		*tp++ = 'A' + unif(0,vr);
	lstchr = *(tp - 1);
	keyl   = unif(vl,fl);
	for (i = vl; i < keyl; i++)
		*tp++ = lstchr;
	for (i = keyl; i < fl; i++)
		*tp++ = ' ';
} /* setval() */


/* ************************************************************************* */
/* ************************************************************************* */

/*
 * getskey: Return a number in the range l..u.
 */

#ifdef PROTOTYPE
COUNT getskey(COUNT l,COUNT u)
#else
COUNT getskey(l,u)
COUNT l,u;
#endif
{
	COUNT retval;

	retval = unif(l,u);
	return(retval);
} /* getskey() */


/* ************************************************************************* */
/* ************************************************************************* */

/*
 * getnum: Return the numeric (LONG) value of a string.
 */

#ifdef PROTOTYPE
LONG getnum(pTEXT tp)
#else
LONG getnum(tp)
pTEXT tp;
#endif
{
	LONG retval;

	retval = 0L;
	do
	{
		if (*tp >= '0' && *tp <= '9')
		{
			retval *= 10;
			retval += (*tp - '0');
		}
	} while (*tp++);
	return(retval);
} /* getnum() */


/* ************************************************************************* */
/* ************************************************************************* */

/*
 * getwrd3: Prompt user for input and get user's input.
 */

#ifdef PROTOTYPE
static VOID getwrd3(pTEXT tp,NINT size,pTEXT prompt)
#else
static VOID getwrd3(tp,size,prompt)
pTEXT tp;
NINT  size;
pTEXT prompt;
#endif
{
	pTEXT sp;
	TEXT  dflt[MAX_NAME];

	sp = tp;
	ctrt_strcpy(dflt,tp); 

loop:
	tp = sp;
#ifdef ctMACINTOSH
	ctrt_printf("\nEnter %s: [%s]\n",prompt,dflt);
	gets(tp);
	while (size)
	{
#else
	ctrt_printf("\nEnter %s: [%s] ",prompt,dflt);
	fflush(stdout);
	while (size)
	{
		*tp = (TEXT)getchar(); /* if possible change to a no echo form */
#endif
		if (*tp == '\r' || *tp == '\n' || *tp == '\0')
		{
			if (sp != tp)
				*tp = '\0';
			else
				*tp = *dflt;
			if (ctrt_strcmp((pTEXT) sp,(pTEXT) "?") == 0)
			{
				ctrt_printf("\n");
				goto loop;
			}
			return;
		}
		tp++;
		size--;
	}
	*(tp - 1) = '\0';
} /* getwrd3() */


/* ************************************************************************* */
/* ************************************************************************* */

/*
 * getstr2: Get a string from user.
 */

#ifdef PROTOTYPE
pTEXT getstr2(pTEXT dflt,pTEXT prompt)
#else
pTEXT getstr2(dflt,prompt)
pTEXT dflt;
pTEXT prompt;
#endif
{
	getwrd3(dflt,MAX_NAME,prompt); 
	return(dflt);
} /* getstr2() */


/* ************************************************************************* */
/* ************************************************************************* */

/*
 * getchr: Get a string from user; return a character.
 */

#ifdef PROTOTYPE
TEXT getchr(TEXT dflt,pTEXT prompt)
#else
TEXT getchr(dflt,prompt)
TEXT  dflt;
pTEXT prompt;
#endif
{
	TEXT inpbuf[MAX_NAME], inpchr;

	ctrt_sprintf(inpbuf,"%c",dflt);
	getwrd3(inpbuf,MAX_NAME,prompt);
	inpchr = inpbuf[0];
	return(inpchr);
} /* getchr() */


/* ************************************************************************* */
/* ************************************************************************* */

/*
 * getlng: Get a string from user; return a LONG.
 */

#ifdef PROTOTYPE
LONG getlng(LONG dflt,pTEXT prompt)
#else
LONG getlng(dflt,prompt)
LONG  dflt;
pTEXT prompt;
#endif
{
	LONG inplng;
	TEXT inpbuf[MAX_NAME];

	ctrt_sprintf(inpbuf, "%ld", dflt);
	getwrd3(inpbuf, MAX_NAME, prompt);
	inplng = getnum(inpbuf);
	return(inplng);
} /* getlng() */


/* ************************************************************************* */
/* ************************************************************************* */

/*
 * getcnt: Get a string from user; return a UCOUNT.
 */

#ifdef PROTOTYPE
UCOUNT getcnt(COUNT dflt,pTEXT prompt)
#else
UCOUNT getcnt(dflt,prompt)
COUNT dflt;
pTEXT prompt;
#endif
{
	TEXT   inpbuf[MAX_NAME];
	UCOUNT inpcnt;

	ctrt_sprintf(inpbuf, "%d", dflt);
	getwrd3(inpbuf, MAX_NAME, prompt);
	inpcnt = (UCOUNT)getnum(inpbuf);
	return(inpcnt);
} /* getcnt() */


/* ************************************************************************* */
/* ************************************************************************* */

/*
 * tcode: Run a trial (executed by each thread).
 */

#ifdef PROTOTYPE
NINT tcode(pVOID t)
#else
NINT tcode(t)
pVOID t;
#endif
{
	NINT		proc_id;
	LONG		round;
	LONG		trials;
	LONG		itrial,lstart,lstop;
	NINT		tlen,trst;
	NINT		status;
	NINT		tcnt = 0;
	COUNT		finds,locks;
	COUNT		key;
	COUNT		datno;
	pAPPOPTS	pOptions;
	DREC		updblk,datblk;
	TEXT		th_str[48];
	TEXT		keybuf[64];
	TEXT		newbuf[64];
	TEXT		qmsg[20];

	pOptions = (pAPPOPTS)t;
	trials   = pOptions->trials;
	proc_id  = (pOptions->round % 100L) - 1;
	round = (pOptions->round / 100L) + 1L;
    	mbfree(pOptions);

	th_printf(proc_id, "Started thread.\n");

	sprintf(th_str, " Round #%ld  Thread #%ld", round, proc_id+1);

	ctThrdSleep(0L);
	lstart = 0L;
	tlen = trst = 0;

	/* Stop the thread. ***************************************************/
	if (action == 'S')
	{
		STPUSR();
		goto exit_thrd;
	}

	if (INTISAMX(16, 12, 16, 16, 0, NULL, NULL, svn))
	{
		th_printf(proc_id, "\nCould not init (%d,%d)\n", isam_err, isam_fil);
		goto exit_thrd;
	}

#ifdef ctCLIENT
	th_printf(proc_id,"Connected to server.\n");
#else
	th_printf(proc_id,"Initialized c-tree Plus.\n");
#endif

	/* Open data and index files. */
	if ((datno = OPNRFILX(-1, "mark.dat", SHARED | DUPCHANEL, NULL)) == -1)
	{
		th_printf(proc_id,"Could not open files (%d,%d)\n",
			isam_err, isam_fil);
		if (isam_err == FNOP_ERR)
			th_printf(proc_id, "Have you created the files?\n");
		goto exit_thrd;
	}

	th_printf(proc_id, "Opened data and index files.\n");

#ifdef TRANPROC
	ctThrdMutexGet(&tfin_mtx);
	if (filmod == (COUNT)-1)
		filmod = (COUNT)GETFIL(datno, FILMOD);
	ctThrdMutexRel(&tfin_mtx);

	if ((COUNT)(filmod & LOGFIL))
	{
		th_printf(proc_id, "Using full transaction processing.\n");
		mode = 'T';
	}
	else if ((COUNT)(filmod & PREIMG))
	{
		th_printf(proc_id, "Using no log transaction processing.\n");
		mode = 't';
	}
	else
	{
		th_printf(proc_id, "Using no transaction processing.\n");
		mode = 'N';
	}
#else
	th_printf(proc_id, "Using no transaction processing.\n");
#endif

	lstart = sysdate();
	th_printf(proc_id,"Start time: %s\n", asctime(ct_tm));
#ifdef ctLOGTIMESTODISK
	fprintf(fpTIMES,"%s  Start time   : %s",th_str,asctime(ct_tm)); 
#endif

	/* Run cttestfunc() routine.*******************************************/
	if (action == 'T')
	{
		for (itrial = 1; itrial <= trials; itrial++)
		{
#ifdef TRANPROC
			if (mode != 'N')
			{
				/* assignments intended */
				status = (tbeg(&tlen,&trst) || cttestfunc());
				if (status)
					th_printf(proc_id,"\n (%ld:%d)\n",itrial,status);
				status = tend(proc_id,&tlen,status,&trst,&tcnt,1);
				if (status)
					th_printf(proc_id,"\n (%ldT%d)\n",itrial,status);
			}
			else
#endif
			{
				status = cttestfunc();
				if (status)
					th_printf(proc_id,"\n (%ld:%d)",itrial,status);
				++tcnt;
				th_printf(proc_id, "\r  TT %d", tcnt);
			}
		}
		goto exit_thrd;
	}

	/* Add data. **********************************************************/
	if (action == 'A')
	{
		th_printf(proc_id,"Adding records...\n");

		datblk.dlfg = 0;
		for (itrial = 1; itrial <= trials; itrial++)
		{
			setval(datblk.unqu,NUML,NUML,27);
			setval(datblk.name,8,NAML,16);
			setval(datblk.clss,CLSL,CLSL,4);
#ifdef TRANPROC
			if (mode != 'N')
			{
				/* assignments intended */
				if ((status=tbeg(&tlen,&trst)) ||
				    (status=ADDREC(datno,&datblk)))
					th_printf(proc_id,"\n (%ld:%d:%.6s)\n",
						itrial,status,datblk.unqu);
				status = tend(proc_id,&tlen,status,&trst,&tcnt,1);
				if (status)
					th_printf(proc_id,"\n (%ldT%d)\n",itrial,status);
				ctThrdSleep(0L);
			}
			else
#endif
			{
		                /* assignments intended */
				if ((status = LKISAM(ENABLE)) ||
				    (status = ADDREC(datno,&datblk)))
					th_printf(proc_id,"\n (%ld:%d:%.6s)\n",
						itrial,status,datblk.unqu);
				ctThrdSleep(0L);
				++tcnt;
				th_printf(proc_id, "\r  TT %d", tcnt);

				/* assignment intended */
				if (status = LKISAM(FREE))
					th_printf(proc_id,"\n (%ldL%d)\n",itrial,status);
				ctThrdSleep(0L);
			}
		}

		goto exit_thrd;
	}

	/* Get, Delete, or Update. ********************************************/
	switch (action)
	{
	case 'G':
		th_printf(proc_id,"Retrieving records...\n");
		break;

	case 'D':
		th_printf(proc_id,"Deleting records...\n");
		break;

	case 'U':
		th_printf(proc_id,"Updating records...\n");
		break;
	}
	
	for (itrial = 1; itrial <= trials; itrial++)
	{
		finds = locks = 0;
gfind:
		for (key = 0; key < 30; key++)
			keybuf[key] = 0;
		key = getskey((COUNT)(datno+1),(COUNT)(datno+2));
		switch (key-datno)
		{
		case 2:
			setval(keybuf,NUML,NUML,27);
		        break;
		case 1:
			setval(keybuf,8,NAML,16);
			break;
		}

#ifdef TRANPROC
		if (mode != 'N')
		/* Transaction processing. *******************************************/
		{
			if (tbeg(&tlen,&trst))
			{
				th_printf(proc_id,"\n (%ld:%d)\n",itrial,isam_err);
				continue;
			}
tlock1:
			status = GTEREC(key,keybuf,&datblk);
			if (status != 0)
			{
				if (status == 101 && finds++ < 10)
				{
					TRANABT();
					tlen = 0;
					goto gfind;
				}
				if (status != 42)
					th_printf(proc_id,"\n (%ld:%d:%d:%s)\n",
						itrial,status,key,keybuf);
				if (status == 42)
				{
					if (locks++ < 10)
						goto tlock1;
					else
						th_printf(proc_id,"\n (%ld:%d)\n",itrial,status);
				}

				/* assignment intended */
				if (status = TRANABT())
					th_printf(proc_id,"\n (%ldT%d)",itrial,status);
				tlen = 0;
				continue;
			}

			switch (action)
			{
case 'G': /* Get. ************************************************************/
			/* Truncate key value for display purposes. */
			if (width >= 6)
				keybuf[width-6] = '\0';
			th_printf(proc_id,"\r G [%s]", keybuf);
			status = tend(proc_id,&tlen,status,&trst,&tcnt,0);
			if (status)
				th_printf(proc_id,"\n (%ldT%d)",itrial,status);
			ctThrdSleep(0);
			break;

case 'D': /* Delete. *********************************************************/
			if ((status = DELREC(datno)) != 0)
				th_printf(proc_id," (D%ld:%d)",itrial,status);
			else
			{
				/* Truncate key value for display purposes. */
				if (width >= 6)
					keybuf[width-6] = '\0';
				th_printf(proc_id,"\r D [%s]", keybuf);
			}
			status = tend(proc_id,&tlen,status,&trst,&tcnt,-1);
			if (status)
				th_printf(proc_id,"\n (%ldT%d)",itrial,status);
			ctThrdSleep(0);
			break;

case 'U': /* Update. *********************************************************/
			cpybuf(&updblk,&datblk,128);
			/* Choose a key to update. */
			switch (getskey((COUNT)(datno+1),(COUNT)(datno+3))-datno)
			{
		case 2:
			cpybuf(keybuf,datblk.unqu,NUML);
			setval(updblk.unqu,NUML,NUML,27);
			cpybuf(newbuf,updblk.unqu,NUML);
			newbuf[NUML] = '\0';
			break;

		case 1:
			cpybuf(keybuf,datblk.name,NAML);
			setval(updblk.name,8,NAML,16);
			cpybuf(newbuf,updblk.name,NAML);
			newbuf[NAML] = '\0';
			break;

		case 3:
			cpybuf(keybuf,datblk.clss,CLSL);
			setval(updblk.clss,CLSL,CLSL,4);
			cpybuf(newbuf,updblk.clss,CLSL);
			newbuf[CLSL] = '\0';
			break;
			}
			if ((status = RWTREC(datno,&updblk)) != 0)
				th_printf(proc_id,"\n (R%ld:%d)",itrial,status);
			else
			{
				/* Truncate key value for display purposes. */
				if (width >= 6)
					newbuf[width-6] = '\0';
				th_printf(proc_id,"\r U [%s]", newbuf);
			}

			status = tend(proc_id,&tlen,status,&trst,&tcnt,0);
			if (status)
				th_printf(proc_id,"\n (%ldT%d)",itrial,status);
			ctThrdSleep(0);
			} /* end switch */
		}
		else
#endif
		/* No transaction processing. ****************************************/
		{
			if (LKISAM(ENABLE))
			{
				th_printf(proc_id,"\n (%ld:%d)\n",itrial,isam_err);
				continue;
			}
tlock0:
			if ((status = GTEREC(key,keybuf,&datblk)) != 0)
			{
				if (status == 101 && finds++ < 10)
				{
					LKISAM(FREE);
					goto gfind;
				}
				if (status != 42 && status != 101)
					th_printf(proc_id,"\n (%ld:%d:%d:%s)\n",
						itrial,status,key,keybuf);
				if (status == 42)
					goto tlock0;
				LKISAM(FREE);
				continue;
			}

			switch (action)
			{
case 'G': /* Get. ************************************************************/
			/* Truncate key value for display purposes. */
			if (width >= 6)
				keybuf[width-6] = '\0';
			th_printf(proc_id,"\r G [%s]", keybuf);
			LKISAM(FREE);
			ctThrdSleep(0);
			break;

case 'D': /* Delete. *********************************************************/
			if ((status = DELREC(datno)) != 0)
				th_printf(proc_id,"\n (%ld:%d)\n",itrial,status);
			else
			{
				/* Truncate key value for display purposes. */
				if (width >= 6)
					keybuf[width-6] = '\0';
				th_printf(proc_id,"\r D [%s]", keybuf);
			}
			LKISAM(FREE);
			ctThrdSleep(0);
			break;

case 'U': /* Update. *********************************************************/
			cpybuf(&updblk,&datblk,128);
			/* Choose a key to update. */
			switch (getskey((COUNT)(datno+1),(COUNT)(datno+3))-datno)
			{
		case 2:
/*
			cpybuf(keybuf,datblk.unqu,NUML);
			setval(updblk.unqu,NUML,NUML,27);
			cpybuf(newbuf,updblk.unqu,NUML);
			newbuf[NUML] = '\0';
			break;
*/
		case 1:
			cpybuf(keybuf,datblk.name,NAML);
			setval(updblk.name,8,NAML,16);
			cpybuf(newbuf,updblk.name,NAML);
			newbuf[NAML] = '\0';
			break;
		case 3:
			cpybuf(keybuf,datblk.clss,CLSL);
			setval(updblk.clss,CLSL,CLSL,4);
			cpybuf(newbuf,updblk.clss,CLSL);
			newbuf[CLSL] = '\0';
			break;
			}
			if ((status = RWTREC(datno,&updblk)) != 0)
				th_printf(proc_id,"\n (%ld:%d)\n",itrial,status);
			else
			{
				/* Truncate key value for display purposes. */
				if (width >= 6)
					newbuf[width-6] = '\0';
				th_printf(proc_id,"\r U [%s]", newbuf);
			}
			LKISAM(FREE);
			ctThrdSleep(0);
			} /* end switch */
		}
	}

exit_thrd:
	if (lstart)
	{
		lstop = sysdate();
		th_printf(proc_id,"\nEnd time: %s", asctime(ct_tm));
#ifdef ctLOGTIMESTODISK
		fprintf(fpTIMES,"%s  End time     : %s ",th_str,asctime(ct_tm)); 
#endif
		rate(proc_id,lstop,lstart,trials,th_str);
		th_printf(proc_id, "***** Successful Run\n\n");
	}

	STPUSR();

	ctThrdMutexGet(&tfin_mtx);
	++tfin;
	ctThrdMutexRel(&tfin_mtx);
	sprintf(qmsg, "Thread %d finished.", proc_id+1);
	ctThrdQueueWrite(qid, qmsg, strlen(qmsg)+1);
	ctThrdExit();

        ctrt_printf(
"ctThrdExit should not return. Call FairCom Tech Support\n");
	exit(1);

	return(0);  
} /* tcode() */


/* ************************************************************************* */
/* ************************************************************************* */

/*
** The following global data and file definitions only to be used by the main
** thread at start-up.
**
** NOTE: global variables and static stack variables are not appropriate for
** use by individual threads unless the variables are not changed by threads,
** and are not meant to be different for each thread. In the event that global
** variables are meant to be changed by individual threads, then a Mutex must
** be used to control access.
*/

DREC	dodablk;	/* Data record buffer.	*/

DATOBJ doda[] = {	/* Field definitions for data record. */
		{"Dlfg", (pTEXT) &dodablk.dlfg, CT_INT2},
		{"Unqu", (pTEXT)  dodablk.unqu, CT_FSTRING, NUML},
		{"Sern", (pTEXT) &dodablk.sern, CT_INT4},
		{"Dlen", (pTEXT) &dodablk.dlen, CT_INT2},
		{"Name", (pTEXT)  dodablk.name, CT_FSTRING, NAML},
		{"Clss", (pTEXT)  dodablk.clss, CT_FSTRING, CLSL},
		{"Grbg", (pTEXT)  dodablk.grbg, CT_FSTRING, PADG}
	};

ISEG segments[] = {	/* Key segment definitions. */
		{14,16,0},
		{ 2, 6,0},
		{38, 2,0},
		{ 8, 4,3}
	};

IIDX ndxs[] = {		/* Index definitions. */
		{20,12,1,1,32,1, segments},
		{10, 0,1,0, 0,1,&segments[1]},
		{10, 0,1,1,32,2,&segments[2]}
	};

IFIL markdata = {	/* Data and index file definitions. */
		"mark",
		-1,			/* file #        */
		128,			/* record length */
		8192,			/* dat xtdsiz	 */
		SHARED | PERMANENT,	/* dat file mode */
		3,			/* nbr of idxs	 */
		0,			/* idx xtdsiz	 */
		SHARED | PERMANENT,	/* idx file mode */
		ndxs			/* idx defs	 */
	};

/*
 * appGetArgs: Process command-line arguments or prompt user.
 */
#ifdef PROTOTYPE
NINT appGetArgs(NINT argc, ppTEXT argv, pAPPOPTS pOptions)
#else
NINT appGetArgs(argc, argv, pOptions)
NINT		argc;
ppTEXT		argv;
pAPPOPTS	pOptions;
#endif
{
	pTEXT	tp;		/* Pointer to current command line argument. */
	NINT	rc;		/* Return code.		*/
	TEXT	fc;		/* File creation flag.	*/

	pOptions->svn = mballc(1,(UINT)MAX_NAME);

	/* DEFAULTS */
	pOptions->action   = 'G';
	pOptions->filestat = 'n';
	pOptions->mode     = 'N';
	pOptions->trials   = 100L;
	ctrt_strcpy(pOptions->svn, SVN_DEFAULT);
	pOptions->seed     = 1;
	pOptions->trnlen   = 1;
	pOptions->maxfiles = 0;
	pOptions->threads  = 1;

	if (argc <= 1) /* if command line args not entered */
	{
		appDescribe();

		pOptions->threads = getcnt((COUNT)pOptions->threads, "Number of Threads");
		pOptions->action  = toupper(getchr(pOptions->action, "Action; A)dd, G)et, D)elete, U)pdate"));
		pOptions->trials  = getlng((LONG)pOptions->trials, "Number of trials"); 
		fc      = getchr(pOptions->filestat, "Create new files? Y)es, N)o");
		if ((fc == 'y') || (fc == 'Y'))
		{
			pOptions->filestat = 'c';
#ifdef TRANPROC
			pOptions->mode    = getchr(pOptions->mode, "Transaction mode; full T)ran, no log t)ran, N)o tran");
			if (pOptions->mode == 'T')
			{
				markdata.dfilmod |= TRNLOG;
				markdata.ifilmod |= TRNLOG;
			}
			else if (pOptions->mode == 't')
			{
				markdata.dfilmod |= PREIMG;
				markdata.ifilmod |= PREIMG;
			}
#endif
		}
		else
		{
			ctrt_printf("\nNote: Will use transaction mode set at file creation time.\n");
			mode = 'D';
		}

		pOptions->seed    = getcnt(pOptions->trnlen, "Random number seed");
#ifdef ctCLIENT
		pOptions->svn     = getstr2(pOptions->svn, "Server name");
#endif

		ctrt_printf("\nParameters are:  threads [%d] action [%c] trials [%ld]\n",
			pOptions->threads, pOptions->action, pOptions->trials);
		ctrt_printf("                 mode [%c] create [%c] seed [%d]\n",
			pOptions->mode, pOptions->filestat, pOptions->seed);
		ctrt_printf("\nPress ENTER to continue...");
		getchar();
	}
	else
	{
	        /* COMMAND LINE */
	        while (--argc > 0)
		{
			tp = *++argv; 
			switch (*tp)
			{
case 'A':
case 'a':		/* Action. */
			pOptions->action = toupper(*++tp);
			break;
case 'F':
case 'f':		/* Create files. */
			pOptions->filestat = *++tp;
			break;
#ifdef TRANPROC
case 'G':
case 'g':		/* Transaction aborts. */
			pOptions->trnlen = atoi(++tp);
			break;
case 'M':
case 'm':		/* Transaction mode. */
			pOptions->mode = *++tp;
			if (pOptions->mode == 'T')
			{
				markdata.dfilmod |= TRNLOG;
				markdata.ifilmod |= TRNLOG;
			}
			else if (pOptions->mode == 't')
			{
				markdata.dfilmod |= PREIMG;
				markdata.ifilmod |= PREIMG;
			}
			break;
#endif
case 'N':
case 'n':		/* Number of trials. */
			pOptions->trials = atol(++tp);
			break;
case 'R':
case 'r':		/* Random number seed. */
			pOptions->seed = atoi(++tp);
			break;
case 'S':
case 's':		/* Server name. */
			pOptions->svn = ++tp;
			break;
case 'X':
case 'x':		/* Maximum number of files. */
			pOptions->maxfiles = atoi(++tp);
			break;
case 'T':
case 't':		/* Number of threads. */
			pOptions->threads = atoi(++tp);
			break;
			}
		}
	}  /* end command line */

#ifdef ctCLIENT
	if (pOptions->maxfiles)
		ctmaxparm = pOptions->maxfiles;
#endif

	if (!strchr("ADGUST", pOptions->action))
	{
		ctrt_printf("Invalid action '%c'\n", pOptions->action);
		goto err_ret;
	}

	ctrt_printf("\ntrials=%ld mode=%c seed=%d action=%c trnlen=%d threads=%d\n",
		pOptions->trials, pOptions->mode, pOptions->seed,
		pOptions->action, pOptions->trnlen, pOptions->threads);

#ifdef ctmarkLOGTODISK
	if ((fpTIMES = fopen("TIMES.LOG","at")) == NULL) 
	{
		ctrt_printf("Cannot open TIMES.LOG file!  Program will now exit!");
		goto err_ret;
	}
	fprintf(fpTIMES,"\ntrials=%ld mode=%c seed=%d action=%c trnlen=%d threads=%d\n\n",
		pOptions->trials, pOptions->mode, pOptions->seed,
		pOptions->action, pOptions->trnlen, pOptions->threads);
#endif

	if (pOptions->threads < 1)
	{
		ctrt_printf("The minimum number of threads is 1.\n");
		goto err_ret;
	}

	if (pOptions->threads > MAXTRD)
	{
		ctrt_printf("\n\n  NOTE: You have chosen to run %d threads;\
 all %d threads will run but only\n", pOptions->threads, pOptions->threads);
		ctrt_printf("        %d threads will be shown on-screen.\n", MAXTRD);
		ctrt_printf("\nPress ENTER to continue...");
		getchar();
	}

	/* Create data and index files if specified. */
	if (pOptions->filestat == 'c' || pOptions->filestat == 'C')
	{
		if (INTISAMX(16, 12, 16, 16, 0, NULL, NULL, pOptions->svn))
		{
			ctrt_printf("\nCould not init (%d,%d)\n", isam_err, isam_fil);
			goto err_ret;
		}

		/* Attempt to delete existing data and index files. */
		markdata.dfilmod &= ~(NONEXCLUSIVE);
		markdata.ifilmod &= ~(NONEXCLUSIVE);
		if (!OPNIFILX(&markdata, NULL, NULL, NULL))
			if (!DELIFIL(&markdata))
				ctrt_printf("\nDeleted existing data and index files.\n");
		markdata.dfilmod |= SHARED;
		markdata.ifilmod |= SHARED;

		/* Create new data and index files. */
		if (rc = CREIFILX(&markdata, NULL, NULL, 0L, NULL, NULL))
		{
			ctrt_printf("\nCould not create %s (%d,%d)\n",
				markdata.pfilnam, isam_err, isam_fil);
			goto err_ret;
		}

		if (PUTDODA(markdata.tfilno, doda, (UCOUNT)7))
			ctrt_printf("\nError on PUTDODA (%d)", uerr_cod);

		STPUSR();
	}

	/* Set globals. */
	mode	= pOptions->mode;
	svn	= pOptions->svn;
	trnlen	= pOptions->trnlen;
	seed	= pOptions->seed;
	action	= pOptions->action;
#ifdef TRANPROC
	filmod  = (COUNT)-1;
#endif

	return (0);

err_ret:
	return (1);
} /* appGetArgs() */


/* ************************************************************************* */
/* ************************************************************************* */

/*
 * appThrdCreate: Create threads which execute the tcode() function.
 */

#ifdef PROTOTYPE
NINT appThrdCreate(pAPPOPTS pOptions)
#else
NINT appThrdCreate(pOptions)
pAPPOPTS pOptions;
#endif
{
	NINT	nthrd = 0;
	NINT	i;

	for (i=0; i<pOptions->threads; i++)
	{
		pAPPOPTS	pOptions2 = (pAPPOPTS)mballc(1,ctSIZE(APPOPTS));
		cpybuf(pOptions2, pOptions, ctSIZE(APPOPTS));
		pOptions2->round = (LONG)(i+1 + 100*(pOptions->round-1));

		if ((pOptions->round == 3) && !i)
		{
			/* Initialize counter which tracks number of completed threads. */
			ctThrdMutexGet(&tfin_mtx);
			tfin = 0;
			ctThrdMutexRel(&tfin_mtx);
		}

		if (ctThrdCreate(tcode,(pVOID)pOptions2,(pVOID)0,(LONG)0) == NO_ERROR)
		{
			++nthrd;
#ifdef WTMTAP_C
			ctrt_printf("\tappThrdCreate: Created thread #%d.\n", i+1);
#endif /* WTMTAP_C */
		}
	}

	return (nthrd);
} /* appThrdCreate() */


/* ************************************************************************* */
/* ************************************************************************* */

/*
 * appThrdWait: Wait on c-tree Plus threads to complete.
 */
#ifdef PROTOTYPE
NINT appThrdWait(pAPPOPTS pOptions, NINT nthrd)
#else
NINT appThrdWait(pOptions, nthrd)
pAPPOPTS	pOptions;
NINT		nthrd;
#endif
{
	NINT	qlen;		/* Number of characters in queue. */
	NINT	rc;		/* Return code. */
	TEXT	qbuf[20];	/* To read in queue contents. */   

	if (pOptions->round != 3)
	{
#ifdef WTMTAP_C
		ctrt_printf("\tappThrdWait: Using inter-thread FIFO queues for notification.\n");
#endif /* WTMTAP_C */
		/* Use inter-thread FIFO queues to wait for all threads to finish. */
		/* Each thread places a message in the queue before exiting. */
		while (nthrd)		
		{				/* FairCom MT API: Please note #04 */
			/* Wait indefinitely for a message; get its length. */
#ifdef WTMTAP_C
			ctrt_printf("\tappThrdWait: Waiting for message to arrive in queue.\n");
#endif /* WTMTAP_C */
			qlen = ctThrdQueueMlen(qid,ctWAITFOREVER);

			/* Retrieve the message we waited for. */
			*qbuf = '\0';
			rc = ctThrdQueueRead(qid,qbuf,19,ctNOWAIT);
#ifdef WTMTAP_C
			if (rc)
				ctrt_printf("\tappThrdWait: Unable to read queue, error = %d\n", rc);
			else
				ctrt_printf("\tappThrdWait: Read queue entry: length=%d, contents=[%s]\n",
					qlen, qbuf);
#else  /* ~WTMTAP_C */
			if (rc)
				th_printf(0,"Unable to read queue, error = %d\n", rc);
			else
				th_printf(0,"Read queue entry: [%s]\n\n", qbuf);
#endif /* ~WTMTAP_C */
			--nthrd;
		}
	}
	else
	{
#ifdef WTMTAP_C
		ctrt_printf("\tappThrdWait: Using global variable and mutex for notification.\n");
#endif
		/* Use value of tfin to wait for all threads to finish. */
		/* Each thread increments tfin before it exits. */
		do
		{
				/* FairCom MT API: Please note #05 */
			ctThrdMutexGet(&tfin_mtx);
			rc = tfin;
			ctThrdMutexRel(&tfin_mtx);
			if (rc >= nthrd)
				break;
#ifdef WTMTAP_C
			ctrt_printf("\tappThrdWait: Number of threads finished: %d\n", rc);
#endif /* WTMTAP_C */
		} while (!ctThrdSleep((LONG) 1000));
#ifdef WTMTAP_C
		ctrt_printf("\tappThrdWait: Number of threads finished: %d\n", rc);
#endif /* WTMTAP_C */
	}

	return (0);
} /* appThrdWait() */


/* end ctmtap.h */
